@@ -6,7 +6,7 @@ import types, time
 from itertools import count
 
 from zope.interface import implements
-from twisted.python import log
+from twisted.python import failure
 from twisted.internet import defer, error
 from twisted.internet import interfaces as twinterfaces
 from twisted.internet.protocol import connectionDone
@@ -18,6 +18,7 @@ from foolscap.tokens import Violation, BananaError
 from foolscap.ipb import DeadReferenceError, IBroker
 from foolscap.slicers.root import RootSlicer, RootUnslicer, ScopedRootSlicer
 from foolscap.eventual import eventually
+from foolscap.logging import log
 
 
 PBTopRegistry = {
@@ -148,6 +149,7 @@ class Broker(banana.Banana, referenceable.Referenceable):
     def __init__(self, remote_tubref, params={},
                  keepaliveTimeout=None, disconnectTimeout=None):
         banana.Banana.__init__(self, params)
+        self._expose_remote_exception_types = True
         self.remote_tubref = remote_tubref
         self.keepaliveTimeout = keepaliveTimeout
         self.disconnectTimeout = disconnectTimeout
@@ -159,7 +161,6 @@ class Broker(banana.Banana, referenceable.Referenceable):
         self.initBroker()
         self.current_slave_IR = params.get('current-slave-IR')
         self.current_seqnum = params.get('current-seqnum')
-        self.current_attempt_id = params.get('current-attempt-id')
         self.creation_timestamp = time.time()
 
     def initBroker(self):
@@ -192,6 +193,7 @@ class Broker(banana.Banana, referenceable.Referenceable):
         assert ipb.ITub.providedBy(tub)
         self.tub = tub
         self.unsafeTracebacks = tub.unsafeTracebacks
+        self._expose_remote_exception_types = tub._expose_remote_exception_types
         if tub.debugBanana:
             self.debugSend = True
             self.debugReceive = True
@@ -216,7 +218,8 @@ class Broker(banana.Banana, referenceable.Referenceable):
 
     def connectionTimedOut(self):
         err = error.ConnectionLost("banana timeout: connection dropped")
-        self.shutdown(err)
+        why = failure.Failure(err)
+        self.shutdown(why)
 
     def shutdown(self, why, fireDisconnectWatchers=True):
         """Stop using this connection. If fireDisconnectWatchers is False,
@@ -226,17 +229,23 @@ class Broker(banana.Banana, referenceable.Referenceable):
         connection quickly, rather than waiting for the transmit queue to
         drain.
         """
+        assert isinstance(why, failure.Failure)
         if not fireDisconnectWatchers:
             self.disconnectWatchers = []
         self.finish(why)
         self.transport.loseConnection()
 
     def connectionLost(self, why):
+        tubid = "?"
+        if self.remote_tubref:
+            tubid = self.remote_tubref.getShortTubID()
+        log.msg("connection to %s lost" % tubid, facility="foolscap.connection")
         self.finish(why)
 
     def finish(self, why):
         if self.disconnected:
             return
+        assert isinstance(why, failure.Failure), why
         self.disconnected = True
         self.remote_broker = None
         self.abandonAllRequests(why)
@@ -281,10 +290,6 @@ class Broker(banana.Banana, referenceable.Referenceable):
         if marker in self.disconnectWatchers:
             self.disconnectWatchers.remove(marker)
 
-    # methods to handle RemoteInterfaces
-    def getRemoteInterfaceByName(self, name):
-        return remoteinterface.RemoteInterfaceRegistry[name]
-
     # methods to send my Referenceables to the other side
 
     def getTrackerForMyReference(self, puid, obj):
@@ -379,8 +384,9 @@ class Broker(banana.Banana, referenceable.Referenceable):
             # release the tracker
             d.addCallback(self.freeYourReferenceTracker, tracker)
         except:
-            log.msg("failure during freeRemoteReference")
-            log.err()
+            f = failure.Failure()
+            log.msg("failure during freeRemoteReference", facility="foolscap",
+                    level=log.UNUSUAL, failure=f)
 
     def freeYourReferenceTracker(self, res, tracker):
         if tracker.received_count != 0:
@@ -478,6 +484,10 @@ class Broker(banana.Banana, referenceable.Referenceable):
             raise Violation("non-existent reqID '%d'" % reqID)
 
     def abandonAllRequests(self, why):
+        # map all connection-lost errors to DeadReferenceError, so
+        # application code only needs to check for one exception type
+        if why.check(error.ConnectionLost, error.ConnectionDone):
+            why = failure.Failure(DeadReferenceError("Connection was lost"))
         for req in self.waitingForAnswers.values():
             eventually(req.fail, why)
 
@@ -502,6 +512,8 @@ class Broker(banana.Banana, referenceable.Referenceable):
         eventually(self.doNextCall)
 
     def doNextCall(self):
+        if self.disconnected:
+            return
         if self._waiting_for_call_to_be_ready:
             return
         if not self.inboundDeliveryQueue:
@@ -585,8 +597,9 @@ class Broker(banana.Banana, referenceable.Referenceable):
             # TODO: .send should return a Deferred that fires when the last
             # byte has been queued, and we should delete the local note then
         except:
-            log.msg("Broker._callfinished unable to send")
-            log.err()
+            f = failure.Failure()
+            log.msg("Broker._callfinished unable to send",
+                    facility="foolscap", level=log.UNUSUAL, failure=f)
         del self.activeLocalCalls[reqID]
 
     def callFailed(self, f, reqID, delivery=None):
@@ -679,7 +692,7 @@ class StorageBroker(Broker):
 # we use it for real, not just for testing. The IConsumer stuff hasn't been
 # tested at all.
 
-class _LoopbackAddress(object):
+class LoopbackAddress(object):
     implements(twinterfaces.IAddress)
 
 class LoopbackTransport(object):
@@ -715,9 +728,9 @@ class LoopbackTransport(object):
         self.protocol.connectionLost(reason)
 
     def getPeer(self):
-        return _LoopbackAddress()
+        return LoopbackAddress()
     def getHost(self):
-        return _LoopbackAddress()
+        return LoopbackAddress()
 
     # IConsumer
     def registerProducer(self, producer, streaming):
