@@ -7,10 +7,11 @@ except ImportError:
     pass
 from zope.interface import implements
 from twisted.internet import reactor, utils, defer
-from twisted.python import usage, procutils, filepath
+from twisted.python import usage, procutils, filepath, log as tw_log
 from twisted.application import service, internet
 import foolscap
 from foolscap.logging.interfaces import RILogGatherer, RILogObserver
+from foolscap.logging.incident import IncidentClassifierBase
 from foolscap.util import get_local_ip_for
 
 class BadTubID(Exception):
@@ -309,6 +310,8 @@ class IncidentObserver(foolscap.Referenceable):
         self.publisher = publisher
         self.stdout = stdout
         self.caught_up_d = defer.Deferred()
+        self.incidents_wanted = []
+        self.incident_fetch_outstanding = False
 
     def connect(self):
         # look for a local state file, to see what incidents we've already
@@ -331,15 +334,30 @@ class IncidentObserver(foolscap.Referenceable):
         return d
 
     def remote_new_incident(self, name, trigger):
-        print >>self.stdout, "got incident", name
+        print >>self.stdout, "new incident", name
         # name= should look like "incident-2008-07-29-204211-aspkxoi". We
         # prevent name= from containing path metacharacters like / or : by
         # using FilePath later on.
+        self.incidents_wanted.append( (name, trigger) )
+        self.maybe_fetch_incident()
+
+    def maybe_fetch_incident(self):
+        # only fetch one incident at a time, to keep the sender's outbound
+        # memory usage to a reasonable level
+        if self.incident_fetch_outstanding:
+            return
+        if not self.incidents_wanted:
+            return
+        self.incident_fetch_outstanding = True
+        (name, trigger) = self.incidents_wanted.pop(0)
+        print >>self.stdout, "fetching incident", name
         d = self.publisher.callRemote("get_incident", name)
         d.addCallback(self._got_incident, name, trigger)
-        d.addCallback(lambda res: None)
-        return d
+        d.addErrback(tw_log.err,
+                     "IncidentObserver.get_incident or _got_incident")
+
     def _got_incident(self, incident, name, trigger):
+        self.incident_fetch_outstanding = False
         # We always save the incident to a .bz2 file.
         abs_fn = self.basedir.child(name).path # this prevents evil
         abs_fn += ".flog.bz2"
@@ -349,6 +367,7 @@ class IncidentObserver(foolscap.Referenceable):
         self.save_incident(abs_fn, incident)
         self.update_latest(name)
         self.gatherer.new_incident(abs_fn, rel_fn, self.tubid_s, incident)
+        self.maybe_fetch_incident()
 
     def save_incident(self, filename, incident):
         now = time.time()
@@ -372,7 +391,7 @@ class IncidentObserver(foolscap.Referenceable):
         self.caught_up_d.callback(None)
         return None
 
-class IncidentGathererService(GatheringBase):
+class IncidentGathererService(GatheringBase, IncidentClassifierBase):
     # create this with 'flogtool create-incident-gatherer BASEDIR'
     # run this as 'cd BASEDIR && twistd -y gatherer.tac'
 
@@ -401,12 +420,10 @@ class IncidentGathererService(GatheringBase):
 
     def __init__(self, classifiers=[], basedir=None, stdout=None):
         GatheringBase.__init__(self, basedir)
-        self.classifiers = []
+        IncidentClassifierBase.__init__(self)
         self.classifiers.extend(classifiers)
         self.stdout = stdout
-
-    def add_classifier(self, f):
-        self.classifiers.append(f)
+        self.incidents_received = 0 # for tests
 
 
     def startService(self):
@@ -416,35 +433,35 @@ class IncidentGathererService(GatheringBase):
         outputdir = os.path.join(self.basedir, "classified")
         if not os.path.isdir(outputdir):
             os.makedirs(outputdir)
-            self.classify_stored_incidents(indir)
+        self.add_classify_files(self.basedir)
+        self.classify_stored_incidents(indir)
         GatheringBase.startService(self)
 
     def classify_stored_incidents(self, indir):
         stdout = self.stdout or sys.stdout
-        print >>stdout, "No classified/ directory: reclassifying stored incidents"
-        # now classify all stored incidents
+        print >>stdout, "classifying stored incidents"
+        # now classify all stored incidents that aren't already classified
+        already = set()
+        outputdir = os.path.join(self.basedir, "classified")
+        for category in os.listdir(outputdir):
+            for line in open(os.path.join(outputdir, category), "r"):
+                fn = line.strip()
+                abs_fn = os.path.join(self.basedir, fn)
+                already.add(abs_fn)
+        print >>stdout, "%d incidents already classified" % len(already)
+        count = 0
         for tubid_s in os.listdir(indir):
             nodedir = os.path.join(indir, tubid_s)
             for fn in os.listdir(nodedir):
                 if fn.startswith("incident-"):
                     abs_fn = os.path.join(nodedir, fn)
+                    if abs_fn in already:
+                        continue
                     incident = self.load_incident(abs_fn)
                     rel_fn = os.path.join("incidents", tubid_s, fn)
-                    self.classify_incident(rel_fn, tubid_s, incident)
-
-    def load_incident(self, abs_fn):
-        assert abs_fn.endswith(".bz2")
-        f = bz2.BZ2File(abs_fn, "r")
-        header = pickle.load(f)["header"]
-        events = []
-        while True:
-            try:
-                wrapped = pickle.load(f)
-            except (EOFError, ValueError):
-                break
-            events.append(wrapped["d"])
-        f.close()
-        return (header, events)
+                    self.move_incident(rel_fn, tubid_s, incident)
+                    count += 1
+        print >>stdout, "done classifying %d stored incidents" % count
 
     def remote_logport(self, nodeid, publisher):
         # we ignore nodeid (which is a printable string), and get the tubid
@@ -460,24 +477,19 @@ class IncidentGathererService(GatheringBase):
 
     def new_incident(self, abs_fn, rel_fn, tubid_s, incident):
         stdout = self.stdout or sys.stdout
-        print >>stdout, "NEW INCIDENT", rel_fn
-        self.classify_incident(rel_fn, tubid_s, incident)
+        self.move_incident(rel_fn, tubid_s, incident)
+        self.incidents_received += 1
 
-    def classify_incident(self, rel_fn, tubid_s, incident):
-        categories = set()
-        for f in self.classifiers:
-            c = f(tubid_s, incident)
-            if c: # allow the classifier to return None, or [], or ["foo"]
-                if isinstance(c, str):
-                    c = [c] # or just "foo"
-                categories.update(c)
-        if not categories:
-            categories.add("unknown")
+    def move_incident(self, rel_fn, tubid_s, incident):
+        stdout = self.stdout or sys.stdout
+        categories = self.classify_incident(incident)
         for c in categories:
             fn = os.path.join(self.basedir, "classified", c)
             f = open(fn, "a")
             f.write(rel_fn + "\n")
             f.close()
+        print >>stdout, "classified %s as [%s]" % (rel_fn, ",".join(categories))
+        return categories
 
 
 INCIDENT_GATHERER_TACFILE = """\
@@ -499,17 +511,19 @@ from twisted.application import service
 
 gs = gatherer.IncidentGathererService()
 
-# To add a classifier function, call gs.add_classifier(f) like this:
+# To add a classifier function, store it in a neighboring file named
+# classify_*.py, in a function named classify_incident(). All such files will
+# be loaded at startup:
 #
-#import re
-#TUBCON_RE = re.compile(r'^Tub.connectorFinished: WEIRD, <foolscap.negotiate.TubConnector instance at \w+> is not in \[')
-#def is_foolscap(nodeid_s, incident):
-#    (header, events) = incident
-#    trigger = header['trigger']
-#    m = trigger.get('message', '')
-#    if TUBCON_RE.search(m):
-#        return 'foolscap-tubconnector'
-#gs.add_classifier(is_foolscap)
+# %% cat classify_foolscap.py
+# import re
+# TUBCON_RE = re.compile(r'^Tub.connectorFinished: WEIRD, <foolscap.negotiate.TubConnector instance at \w+> is not in \[')
+# def classify_incident(trigger):
+#     # match some foolscap messages
+#     m = trigger.get('message', '')
+#     if TUBCON_RE.search(m):
+#         return 'foolscap-tubconnector'
+# %%
 
 application = service.Application('incident_gatherer')
 gs.setServiceParent(application)
