@@ -15,7 +15,7 @@ from foolscap.reconnector import Reconnector
 from foolscap.logging import log as flog
 from foolscap.logging import log
 from foolscap.logging import publish as flog_publish
-from foolscap.logging.log import WEIRD, UNUSUAL
+from foolscap.logging.log import UNUSUAL
 
 crypto_available = False
 try:
@@ -147,6 +147,9 @@ class Listener(protocol.ServerFactory):
     def lookupTubID(self, tubID):
         return self.tubs.get(tubID), self.redirects.get(tubID)
 
+def generateSwissnumber(bits):
+    bytes = os.urandom(bits/8)
+    return base32.encode(bytes)
 
 class Tub(service.MultiService):
     """I am a presence in the PB universe, also known as a Tub.
@@ -305,8 +308,10 @@ class Tub(service.MultiService):
 
         self._log_gatherer_furl = None
         self._log_gatherer_furlfile = None
+        self._log_gatherer_connectors = {} # maps furl to reconnector
 
         self._handle_old_duplicate_connections = False
+        self._expose_remote_exception_types = True
 
     def setOption(self, name, value):
         if name == "logLocalFailures":
@@ -351,16 +356,18 @@ class Tub(service.MultiService):
             if value is True:
                 value = 60
             self._handle_old_duplicate_connections = int(value)
+        elif name == "expose-remote-exception-types":
+            self._expose_remote_exception_types = bool(value)
         else:
             raise KeyError("unknown option name '%s'" % name)
 
     def setLogGathererFURL(self, gatherer_furl):
-        assert not self._log_gatherer_furl and not self._log_gatherer_furlfile
+        assert not self._log_gatherer_furl
         self._log_gatherer_furl = gatherer_furl
         self._maybeConnectToGatherer()
 
     def setLogGathererFURLFile(self, gatherer_furlfile):
-        assert not self._log_gatherer_furl and not self._log_gatherer_furlfile
+        assert not self._log_gatherer_furlfile
         self._log_gatherer_furlfile = gatherer_furlfile
         self._maybeConnectToGatherer()
 
@@ -380,7 +387,10 @@ class Tub(service.MultiService):
             except EnvironmentError:
                 pass
         for f in furls:
+            if f in self._log_gatherer_connectors:
+                continue
             connector = self.connectTo(f, self._log_gatherer_connected)
+            self._log_gatherer_connectors[f] = connector
 
     def _log_gatherer_connected(self, rref):
         # we want the logport's furl to be nailed down now, so we'll use the
@@ -575,20 +585,10 @@ class Tub(service.MultiService):
         assert self.running
         self._activeConnectors.append(c)
     def connectorFinished(self, c):
-        if c not in self._activeConnectors:
-            # TODO: I've seen this happen, but I can't figure out how it
-            # could possibly happen. Log and ignore rather than exploding
-            # when we try to do .remove, since this whole connector-tracking
-            # thing is mainly for the benefit of the unit tests (applications
-            # which never shut down a Tub aren't going to care), and it is
-            # more important to let application code run normally than to
-            # force an error here.
-            self.log("Tub.connectorFinished: WEIRD, %s is not in %s"
-                     % (c, self._activeConnectors), level=WEIRD)
-            return
-        self._activeConnectors.remove(c)
-        if not self.running and not self._activeConnectors:
-            self._allConnectorsAreFinished.fire(self)
+        if c in self._activeConnectors:
+            self._activeConnectors.remove(c)
+            if not self.running and not self._activeConnectors:
+                self._allConnectorsAreFinished.fire(self)
 
     def startService(self):
         service.MultiService.startService(self)
@@ -642,8 +642,7 @@ class Tub(service.MultiService):
         return defer.DeferredList(dl)
 
     def generateSwissnumber(self, bits):
-        bytes = os.urandom(bits/8)
-        return base32.encode(bytes)
+        return generateSwissnumber(bits)
 
     def buildURL(self, name):
         if self.encrypted:
@@ -824,9 +823,13 @@ class Tub(service.MultiService):
         The Tub must be running (i.e. Tub.startService()) when this is
         invoked. Future releases may relax this requirement.
 
-        @return: a Deferred that fires with the RemoteReference
+        @return: a Deferred that fires with the RemoteReference. Any failures
+        are returned asynchronously.
         """
 
+        return defer.maybeDeferred(self._getReference, sturdyOrURL)
+
+    def _getReference(self, sturdyOrURL):
         if isinstance(sturdyOrURL, SturdyRef):
             sturdy = sturdyOrURL
         else:
