@@ -5,6 +5,7 @@
 # live in call.py
 
 import weakref
+import re
 
 from zope.interface import interface
 from zope.interface import implements
@@ -13,7 +14,7 @@ Interface = interface.Interface
 from twisted.internet import defer
 from twisted.python import failure, log
 
-from foolscap import ipb, slicer, tokens, call
+from foolscap import ipb, slicer, tokens, call, base32
 BananaError = tokens.BananaError
 Violation = tokens.Violation
 from foolscap.constraint import IConstraint, ByteStringConstraint
@@ -61,7 +62,7 @@ class Referenceable(OnlyReferenceable):
 
 constraintMap[Referenceable] = RemoteInterfaceConstraint(None)
 
-class ReferenceableTracker:
+class ReferenceableTracker(object):
     """I hold the data which tracks a local Referenceable that is in used by
     a remote Broker.
 
@@ -207,8 +208,8 @@ class ReferenceUnslicer(slicer.BaseUnslicer):
     clid = None
     interfaceName = None
     url = None
-    inameConstraint = ByteStringConstraint(200) # TODO: only known RI names?
-    urlConstraint = ByteStringConstraint(200)
+    inameConstraint = ByteStringConstraint() # TODO: only known RI names?
+    urlConstraint = ByteStringConstraint()
 
     def checkToken(self, typebyte, size):
         if self.state == 0:
@@ -255,7 +256,7 @@ class ReferenceUnslicer(slicer.BaseUnslicer):
 
 
 
-class RemoteReferenceTracker:
+class RemoteReferenceTracker(object):
     """I hold the data necessary to locate (or create) a RemoteReference.
 
     @ivar url: the target Referenceable's global URL
@@ -277,13 +278,21 @@ class RemoteReferenceTracker:
         # TODO: the remote end sends us a global URL, when really it should
         # probably send us a per-Tub name, which can can then concatenate to
         # their TubID if/when we pass it on to others. By accepting a full
-        # URL, we give them the ability to sort-of spoof others. We could
-        # check that url.startswith(broker.remoteTub.baseURL), but the Right
-        # Way is to just not have them send the base part in the first place.
-        # I haven't yet made this change because I'm not yet positive it
-        # would work.. how exactly does the base url get sent, anyway? What
-        # about Tubs visible through multiple names?
+        # URL, we give them the ability to sort-of spoof others. For now, we
+        # check that their URL uses the same tubid as our broker is
+        # expecting, but the Right Way is to just not have them send the base
+        # part in the first place. I haven't yet made this change because I'm
+        # not yet positive it would work.. how exactly does the base url get
+        # sent, anyway? What about Tubs visible through multiple names?
         self.url = url
+        if url is not None and self.broker.remote_tubref:
+            # unit tests frequently set url=None, and UnauthenticatedTubs
+            # don't have a remote_tubref
+            expected_tubid = self.broker.remote_tubref.getTubID()
+            url_tubid = SturdyRef(url).getTubRef().getTubID()
+            if expected_tubid != url_tubid:
+                raise BananaError("inbound reference claims bad tubid, %s vs %s"
+                                  % (expected_tubid, url_tubid))
         self.interfaceName = interfaceName
         self.interface = getRemoteInterfaceByName(interfaceName)
         self.received_count = 0
@@ -338,6 +347,26 @@ class RemoteReferenceOnly(object):
 
     def getSturdyRef(self):
         return SturdyRef(self.tracker.getURL())
+    def getRemoteTubID(self):
+        rt = self.tracker.broker.remote_tubref
+        if rt:
+            return rt.getTubID()
+        return "<unauth>"
+
+    def getPeer(self):
+        """Return an IAddress-providing object that describes the remote
+        peer. If we've connected to ourselves, this will be a
+        foolscap.broker.LoopbackAddress instance. If we've connected to
+        someone else, this will be a twisted.internet.address.IPv4Address
+        instance, with .host and .port attributes."""
+        transport = self.tracker.broker.transport
+        return transport.getPeer()
+
+    def isConnected(self):
+        """Return False if this reference is known to be dead."""
+        return not self.tracker.broker.disconnected
+    def getLocationHints(self):
+        return SturdyRef(self.tracker.url).locationHints
 
     def notifyOnDisconnect(self, callback, *args, **kwargs):
         """Register a callback to run when we lose this connection.
@@ -565,7 +594,7 @@ class RemoteMethodReference(RemoteReference):
         methodSchema = None
         return interfaceName, methodName, methodSchema
 
-class LocalReferenceable:
+class LocalReferenceable(object):
     implements(ipb.IRemoteReference)
     def __init__(self, original):
         self.original = original
@@ -654,7 +683,7 @@ class TheirReferenceUnslicer(slicer.LeafUnslicer):
     state = 0
     giftID = None
     url = None
-    urlConstraint = ByteStringConstraint(200)
+    urlConstraint = ByteStringConstraint()
 
     def checkToken(self, typebyte, size):
         if self.state == 0:
@@ -726,6 +755,74 @@ class TheirReferenceUnslicer(slicer.LeafUnslicer):
             return "<gift-?>"
         return "<gift-%s>" % self.giftID
 
+AUTH_STURDYREF_RE = re.compile(r"pb://([^@]+)@([^/]+)/(.+)$")
+NONAUTH_STURDYREF_RE = re.compile(r"pbu://([^/]+)/(.+)$")
+
+IPV4_HINT_RE = re.compile(r"^([^:]+):(\d+)$")
+def encode_location_hint(hint):
+    assert hint[0] == "ipv4"
+    host, port = hint[1:]
+    return "%s:%d" % (host, port)
+def decode_location_hints(hints_s):
+    hints = []
+    for hint_s in hints_s.split(","):
+        mo = IPV4_HINT_RE.search(hint_s)
+        if mo:
+            hint = ( "ipv4", mo.group(1), int(mo.group(2)) )
+            hints.append(hint)
+        else:
+            # some extension from the future that we will ignore
+            pass
+    return hints
+
+class BadFURLError(Exception):
+    pass
+
+def decode_furl(furl):
+    """Returns (encrypted, tubID, location_hints, name)"""
+    # pb://key@{ip:port,host:port,[ipv6]:port}[/unix]/swissnumber
+    # i.e. pb://tubID@{locationHints..}/name
+    #
+    # it can live at any one of a (TODO) variety of network-accessible
+    # locations, or (TODO) at a single UNIX-domain socket.
+    #
+    # there is also an unauthenticated form, which does not have a TubID
+
+    mo_auth_furl = AUTH_STURDYREF_RE.search(furl)
+    mo_nonauth_furl = NONAUTH_STURDYREF_RE.search(furl)
+    if mo_auth_furl:
+        encrypted = True
+        # we only pay attention to the first 32 base32 characters
+        # of the tubid string. Everything else is left for future
+        # extensions.
+        tubID_s = mo_auth_furl.group(1)
+        tubID = tubID_s[:32]
+        if not base32.is_base32(tubID):
+            raise BadFURLError("'%s' is not a valid tubid" % (tubID,))
+        hints = mo_auth_furl.group(2)
+        location_hints = decode_location_hints(hints)
+        name = mo_auth_furl.group(3)
+
+    elif mo_nonauth_furl:
+        encrypted = False
+        tubID = None
+        hints = mo_nonauth_furl.group(1)
+        location_hints = decode_location_hints(hints)
+        name = mo_nonauth_furl.group(2)
+
+    else:
+        raise ValueError("unknown FURL prefix in %r" % (furl,))
+    return (encrypted, tubID, location_hints, name)
+
+def encode_furl(encrypted, tubID, location_hints, name):
+    location_hints_s = ",".join([encode_location_hint(hint)
+                                 for hint in location_hints])
+    if encrypted:
+        return "pb://" + tubID + "@" + location_hints_s + "/" + name
+    else:
+        return "pbu://" + location_hints + "/" + name
+
+
 class SturdyRef(Copyable, RemoteCopy):
     """I am a pointer to a Referenceable that lives in some (probably remote)
     Tub. This pointer is long-lived, however you cannot send messages with it
@@ -743,64 +840,36 @@ class SturdyRef(Copyable, RemoteCopy):
 
     encrypted = False
     tubID = None
-    location = None
-    locationHints = []
     name = None
 
     def __init__(self, url=None):
+        self.locationHints = [] # list of ("ipv4", host, port) tuples
+        self.url = url
         if url:
-            # pb://key@{ip:port,host:port,[ipv6]:port}[/unix]/swissnumber
-            # i.e. pb://tubID@{locationHints..}/name
-            #
-            # it can live at any one of a variety of network-accessible
-            # locations, or at a single UNIX-domain socket.
-            #
-            # there is also an unauthenticated form, which is indexed by the
-            # single locationHint, because it does not have a TubID
-
-            if url.startswith("pb://"):
-                self.encrypted = True
-                url = url[len("pb://"):]
-                slash = url.rfind("/")
-                self.name = url[slash+1:]
-                at = url.find("@")
-                if at != -1:
-                    self.tubID = url[:at]
-                self.locationHints = url[at+1:slash].split(",")
-            elif url.startswith("pbu://"):
-                self.encrypted = False
-                url = url[len("pbu://"):]
-                slash = url.rfind("/")
-                self.name = url[slash+1:]
-                self.tubID = None
-                self.location = url[:slash]
-            else:
-                raise ValueError("unknown FURL prefix in %r" % (url,))
+            self.encrypted, self.tubID, self.locationHints, self.name = \
+                decode_furl(url)
 
     def getTubRef(self):
         if self.encrypted:
             return TubRef(self.tubID, self.locationHints)
-        return NoAuthTubRef(self.location)
+        return NoAuthTubRef(self.locationHints)
+
 
     def getURL(self):
-        if self.encrypted:
-            return ("pb://" + self.tubID + "@" +
-                    ",".join(self.locationHints) +
-                    "/" + self.name)
-        return "pbu://" + self.location + "/" + self.name
+        return self.url
 
     def __str__(self):
-        return self.getURL()
+        return str(self.url)
 
     def _distinguishers(self):
         """Two SturdyRefs are equivalent if they point to the same object.
         SturdyRefs to encrypted Tubs only pay attention to the TubID and the
         reference name. SturdyRefs to unauthenticated Tubs must use the
-        location hint instead of the (missing) TubID. This method makes it
+        location hints instead of the (missing) TubID. This method makes it
         easier to compare a pair of SturdyRefs."""
         if self.encrypted:
             return (True, self.tubID, self.name)
-        return (False, self.location, self.name)
+        return (False, self.locationHints, self.name)
 
     def __hash__(self):
         return hash(self._distinguishers())
@@ -810,7 +879,7 @@ class SturdyRef(Copyable, RemoteCopy):
                 cmp(self._distinguishers(), them._distinguishers()))
 
 
-class TubRef:
+class TubRef(object):
     """This is a little helper class which provides a comparable identifier
     for Tubs. TubRefs can be used as keys in dictionaries that track
     connections to remote Tubs."""
@@ -846,11 +915,11 @@ class NoAuthTubRef(TubRef):
     # this is only used on outbound connections
     encrypted = False
 
-    def __init__(self, location):
-        self.location = location
+    def __init__(self, locations):
+        self.locations = locations
 
     def getLocations(self):
-        return [self.location]
+        return self.locations
 
     def getTubID(self):
         return "<unauth>"
@@ -858,8 +927,9 @@ class NoAuthTubRef(TubRef):
         return "<unauth>"
 
     def __str__(self):
-        return "pbu://" + self.location
+        return "pbu://" + ",".join([encode_location_hint(location)
+                                    for location in self.locations])
 
     def _distinguishers(self):
         """This serves the same purpose as SturdyRef._distinguishers."""
-        return (self.location,)
+        return tuple(self.locations)
