@@ -2,32 +2,15 @@
 from zope.interface import implements
 from twisted.trial import unittest
 from twisted.internet import defer, protocol, reactor
-from twisted.internet.error import ConnectionDone, ConnectionLost, \
-     ConnectionRefusedError
+from twisted.internet.error import ConnectionRefusedError
 from twisted.python import failure
-from foolscap import Tub, UnauthenticatedTub, RemoteInterface, Referenceable
-from foolscap.referenceable import RemoteReference, SturdyRef
-from foolscap.test.common import HelperTarget, RIHelper
+from foolscap import RemoteInterface, Referenceable
+from foolscap.referenceable import RemoteReference, encode_furl, decode_furl
+from foolscap.test.common import HelperTarget, RIHelper, \
+     crypto_available, GoodEnoughTub
 from foolscap.eventual import flushEventualQueue
 from foolscap.tokens import BananaError, NegotiationError
 
-crypto_available = False
-try:
-    from foolscap import crypto
-    crypto_available = crypto.available
-except ImportError:
-    pass
-
-# we use authenticated tubs if possible. If crypto is not available, fall
-# back to unauthenticated ones
-GoodEnoughTub = UnauthenticatedTub
-if crypto_available:
-    GoodEnoughTub = Tub
-
-def ignoreConnectionDone(f):
-    f.trap(ConnectionDone, ConnectionLost)
-    return None
-
 class RIConstrainedHelper(RemoteInterface):
     def set(obj=RIHelper): return None
 
@@ -51,7 +34,8 @@ class Base:
         for s in self.services:
             s.startService()
             l = s.listenOn("tcp:0:interface=127.0.0.1")
-            s.setLocation("127.0.0.1:%d" % l.getPortnum())
+            loc = "127.0.0.1:%d" % l.getPortnum()
+            s.setLocation(loc)
 
     def tearDown(self):
         d = defer.DeferredList([s.stopService() for s in self.services])
@@ -172,11 +156,8 @@ class Gifts(Base, unittest.TestCase):
             d2 = self.bob.waitfor()
             if self.debug: print "Alice introduces Carol to Bob"
             # send the gift. This might not get acked by the time the test is
-            # done and everything is torn down, so explicitly silence any
-            # ConnectionDone error that might result. When we get
-            # callRemoteOnly(), use that instead.
-            d3 = self.abob.callRemote("set", obj=(self.alice, self.acarol))
-            d3.addErrback(ignoreConnectionDone)
+            # done and everything is torn down, so we use callRemoteOnly
+            self.abob.callRemoteOnly("set", obj=(self.alice, self.acarol))
             return d2 # this fires with the gift that bob got
         d.addCallback(_introduce)
         def _bobGotCarol((balice,bcarol)):
@@ -185,8 +166,7 @@ class Gifts(Base, unittest.TestCase):
             if self.debug: print "Bob says something to Carol"
             d2 = self.carol.waitfor()
             # handle ConnectionDone as described before
-            d3 = self.bcarol.callRemote("set", obj=12)
-            d3.addErrback(ignoreConnectionDone)
+            self.bcarol.callRemoteOnly("set", obj=12)
             return d2
         d.addCallback(_bobGotCarol)
         def _carolCalled(res):
@@ -194,6 +174,7 @@ class Gifts(Base, unittest.TestCase):
             self.failUnlessEqual(res, 12)
         d.addCallback(_carolCalled)
         return d
+    testGift.timeout = 10
 
     def testImplicitGift(self):
         # in this test, Carol was registered in her Tub (using
@@ -222,10 +203,9 @@ class Gifts(Base, unittest.TestCase):
             # done and everything is torn down, so explicitly silence any
             # ConnectionDone error that might result. When we get
             # callRemoteOnly(), use that instead.
-            d3 = self.abob.callRemote("set", obj=(self.alice,
-                                                  self.acarol,
-                                                  a_cindy))
-            d3.addErrback(ignoreConnectionDone)
+            self.abob.callRemoteOnly("set", obj=(self.alice,
+                                                 self.acarol,
+                                                 a_cindy))
             return d2 # this fires with the gift that bob got
         d.addCallback(_introduce)
         def _bobGotCarol((b_alice,b_carol,b_cindy)):
@@ -240,10 +220,8 @@ class Gifts(Base, unittest.TestCase):
             d3 = self.cindy.waitfor()
 
             # handle ConnectionDone as described before
-            d4 = b_carol.callRemote("set", obj=4)
-            d4.addErrback(ignoreConnectionDone)
-            d5 = b_cindy.callRemote("set", obj=5)
-            d5.addErrback(ignoreConnectionDone)
+            b_carol.callRemoteOnly("set", obj=4)
+            b_cindy.callRemoteOnly("set", obj=5)
             return defer.DeferredList([d2,d3])
         d.addCallback(_bobGotCarol)
         def _carolAndCindyCalled(res):
@@ -450,9 +428,7 @@ class Bad(Base, unittest.TestCase):
             # which will result in a failure in remote_getReferenceByName.
             # NOTE: this will have to change when we modify the way gifts are
             # referenced, since tracker.url is scheduled to go away.
-            r = SturdyRef(adave.tracker.url)
-            r.name += ".MANGLED"
-            adave.tracker.url = r.getURL()
+            adave.tracker.url = adave.tracker.url + ".MANGLED"
             return self.acarol.callRemote("set", adave)
         d.addCallback(_introduce)
         d.addBoth(self.shouldFail, KeyError, "Bad.test_swissnum")
@@ -460,7 +436,6 @@ class Bad(Base, unittest.TestCase):
         d.addCallback(lambda res: self.acarol.callRemote("set", 14))
         d.addCallback(lambda res: self.failUnlessEqual(self.carol.obj, 14))
         return d
-    test_swissnum.timeout = 10
 
     def test_tubid(self):
         self.createCharacters()
@@ -468,18 +443,21 @@ class Bad(Base, unittest.TestCase):
         d.addCallback(lambda res: self.tubA.getReference(self.dave_url))
         def _introduce(adave):
             # The second way is to mangle the tubid, which will result in a
-            # failure during negotiation. NOTE: this will have to change when
-            # we modify the way gifts are referenced, since tracker.url is
-            # scheduled to go away.
-            r = SturdyRef(adave.tracker.url)
-            r.tubID += ".MANGLED"
-            adave.tracker.url = r.getURL()
+            # failure during negotiation. We mangle it by reversing the
+            # characters: this makes it syntactically valid but highly
+            # unlikely to remain the same. NOTE: this will have to change
+            # when we modify the way gifts are referenced, since tracker.url
+            # is scheduled to go away.
+            (encrypted, tubid, location_hints, name) = \
+                decode_furl(adave.tracker.url)
+            tubid = "".join(reversed(tubid))
+            adave.tracker.url = encode_furl(encode_furl, tubid,
+                                            location_hints, name)
             return self.acarol.callRemote("set", adave)
         d.addCallback(_introduce)
         d.addBoth(self.shouldFail, BananaError, "Bad.test_tubid",
                   "unknown TubID")
         return d
-    test_tubid.timeout = 10
 
     def test_location(self):
         self.createCharacters()
@@ -489,15 +467,17 @@ class Bad(Base, unittest.TestCase):
             # The third way is to mangle the location hints, which will
             # result in a failure during negotiation as it attempts to
             # establish a TCP connection.
-            r = SturdyRef(adave.tracker.url)
+
+            (encrypted, tubid, location_hints, name) = \
+                decode_furl(adave.tracker.url)
             # highly unlikely that there's anything listening on this port
-            r.locationHints = ["127.0.0.47:1"]
-            adave.tracker.url = r.getURL()
+            location_hints = [ ("ipv4", "127.0.0.1", 2) ]
+            adave.tracker.url = encode_furl(encode_furl, tubid,
+                                            location_hints, name)
             return self.acarol.callRemote("set", adave)
         d.addCallback(_introduce)
         d.addBoth(self.shouldFail, ConnectionRefusedError, "Bad.test_location")
         return d
-    test_location.timeout = 10
 
     def test_hang(self):
         f = protocol.Factory()
@@ -513,9 +493,11 @@ class Bad(Base, unittest.TestCase):
             # case, but we can connect to a port which accepts the connection
             # and then stays silent. This should trigger the overall
             # connection timeout.
-            r = SturdyRef(adave.tracker.url)
-            r.locationHints = ["127.0.0.1:%d" % p.getHost().port]
-            adave.tracker.url = r.getURL()
+            (encrypted, tubid, location_hints, name) = \
+                decode_furl(adave.tracker.url)
+            location_hints = [ ("ipv4", "127.0.0.1", p.getHost().port) ]
+            adave.tracker.url = encode_furl(encode_furl, tubid,
+                                            location_hints, name)
             self.tubD.options['connect_timeout'] = 2
             return self.acarol.callRemote("set", adave)
         d.addCallback(_introduce)
@@ -529,8 +511,7 @@ class Bad(Base, unittest.TestCase):
             return d1
         d.addBoth(_stop_listening)
         return d
-    test_hang.timeout = 10
-    
+
 
     def testReturn_swissnum(self):
         self.createCharacters()
@@ -541,9 +522,7 @@ class Bad(Base, unittest.TestCase):
             # which will result in a failure in remote_getReferenceByName.
             # NOTE: this will have to change when we modify the way gifts are
             # referenced, since tracker.url is scheduled to go away.
-            r = SturdyRef(self.bdave.tracker.url)
-            r.name += ".MANGLED"
-            self.bdave.tracker.url = r.getURL()
+            self.bdave.tracker.url = self.bdave.tracker.url + ".MANGLED"
             self.bob.obj = self.bdave
             return self.abob.callRemote("get")
         d.addCallback(_introduce)
@@ -552,4 +531,41 @@ class Bad(Base, unittest.TestCase):
         d.addCallback(lambda res: self.abob.callRemote("set", 14))
         d.addCallback(lambda res: self.failUnlessEqual(self.bob.obj, 14))
         return d
-    testReturn_swissnum.timeout = 10
+
+class LongFURL(Base, unittest.TestCase):
+    # make sure the old 200-byte limit on gift FURLs is gone
+    def setUp(self):
+        self.services = [GoodEnoughTub() for i in range(4)]
+        self.tubA, self.tubB, self.tubC, self.tubD = self.services
+        for s in self.services:
+            s.startService()
+            l = s.listenOn("tcp:0:interface=127.0.0.1")
+            loc = "127.0.0.1:%d" % l.getPortnum()
+            loc = ",".join([loc]*15) # 239 bytes of location, 281 of FURL
+            s.setLocation(loc)
+
+    def testGift(self):
+        self.createCharacters()
+        d = self.createInitialReferences()
+        def _introduce(res):
+            d2 = self.bob.waitfor()
+            if self.debug: print "Alice introduces Carol to Bob"
+            # send the gift. This might not get acked by the time the test is
+            # done and everything is torn down, so we use callRemoteOnly
+            self.abob.callRemoteOnly("set", obj=(self.alice, self.acarol))
+            return d2 # this fires with the gift that bob got
+        d.addCallback(_introduce)
+        def _bobGotCarol((balice,bcarol)):
+            if self.debug: print "Bob got Carol"
+            self.bcarol = bcarol
+            if self.debug: print "Bob says something to Carol"
+            d2 = self.carol.waitfor()
+            # handle ConnectionDone as described before
+            self.bcarol.callRemoteOnly("set", obj=12)
+            return d2
+        d.addCallback(_bobGotCarol)
+        def _carolCalled(res):
+            if self.debug: print "Carol heard from Bob"
+            self.failUnlessEqual(res, 12)
+        d.addCallback(_carolCalled)
+        return d
