@@ -9,10 +9,10 @@ if False:
 
 from twisted.python import failure, reflect
 from twisted.internet import defer
+from twisted.internet.interfaces import IAddress
 from twisted.trial import unittest
 
 from foolscap import tokens, referenceable
-from foolscap import Tub, UnauthenticatedTub
 from foolscap import getRemoteURL_TCP
 from foolscap.tokens import BananaError, Violation, INT, STRING, OPEN
 from foolscap.tokens import BananaFailure
@@ -20,23 +20,10 @@ from foolscap import broker, call
 from foolscap.constraint import IConstraint
 from foolscap.logging import log
 
-crypto_available = False
-try:
-    from foolscap import crypto
-    crypto_available = crypto.available
-except ImportError:
-    pass
-
-# we use authenticated tubs if possible. If crypto is not available, fall
-# back to unauthenticated ones
-GoodEnoughTub = UnauthenticatedTub
-if crypto_available:
-    GoodEnoughTub = Tub
-
-from foolscap.test.common import HelperTarget, RIHelper, TargetMixin
+from foolscap.test.common import HelperTarget, RIHelper, TargetMixin, \
+     Target, TargetWithoutInterfaces, crypto_available, GoodEnoughTub
 from foolscap.eventual import fireEventually, flushEventualQueue
 
-from foolscap.test.common import Target, TargetWithoutInterfaces
 
 
 class TestRequest(call.PendingRequest):
@@ -223,8 +210,13 @@ class TestReferenceable(TargetMixin, unittest.TestCase):
         d.addCallback(self._testRef1_1, r)
         return d
     def _testRef1_1(self, res, r):
-        t = res.tracker
         self.failUnless(isinstance(res, referenceable.RemoteReference))
+        rref = res
+        self.failUnless(isinstance(rref.getPeer(), broker.LoopbackAddress))
+        self.failUnlessEqual(rref.isConnected(), True)
+        self.failUnlessEqual(rref.getLocationHints(), []) # loopback
+        self.failUnlessEqual(rref.getSturdyRef().getURL(), None)
+        t = rref.tracker
         self.failUnlessEqual(t.broker, self.targetBroker)
         self.failUnless(type(t.clid) is int)
         self.failUnless(self.callingBroker.getMyReferenceByCLID(t.clid) is r)
@@ -401,10 +393,12 @@ class TestCallable(unittest.TestCase):
     def setUp(self):
         self.services = [GoodEnoughTub(), GoodEnoughTub()]
         self.tubA, self.tubB = self.services
+        self.tub_ports = []
         for s in self.services:
             s.startService()
             l = s.listenOn("tcp:0:interface=127.0.0.1")
             s.setLocation("127.0.0.1:%d" % l.getPortnum())
+            self.tub_ports.append(l.getPortnum())
         self._log_observers_to_remove = []
 
     def addLogObserver(self, observer):
@@ -425,6 +419,14 @@ class TestCallable(unittest.TestCase):
         def _check(rref):
             sr = rref.getSturdyRef()
             self.failUnlessEqual(sr.getURL(), url)
+            peer = rref.getPeer()
+            self.failUnless(IAddress.providedBy(peer))
+            self.failUnlessEqual(peer.type, "TCP")
+            self.failUnlessEqual(peer.host, "127.0.0.1")
+            self.failUnlessEqual(rref.getRemoteTubID(), self.tubB.getTubID())
+            self.failUnlessEqual(rref.isConnected(), True)
+            self.failUnlessEqual(rref.getLocationHints(),
+                                 [('ipv4', '127.0.0.1', self.tub_ports[1])])
         d.addCallback(_check)
         return d
 
